
#include <iostream>
#include <vkcv/Buffer.hpp>
#include <vkcv/Core.hpp>
#include <vkcv/DescriptorWrites.hpp>
#include <vkcv/Image.hpp>
#include <vkcv/Pass.hpp>
#include <vkcv/Sampler.hpp>
#include <vkcv/Window.hpp>
#include <vkcv/asset/asset_loader.hpp>
#include <vkcv/camera/CameraManager.hpp>
#include <vkcv/geometry/Cuboid.hpp>
#include <vkcv/geometry/Sphere.hpp>
#include <vkcv/geometry/Teapot.hpp>
#include <vkcv/gui/GUI.hpp>
#include <vkcv/shader/GLSLCompiler.hpp>

struct Material {
	glm::vec3 diffuse;
	float reflectionStrength;
	glm::vec3 specular;
	float shininess;
};

struct Light {
	glm::vec4 pos; // pos.w = 0 (dir light), pos.w = 1 (point light)
	glm::vec3 col;
	float spot_exponent;
	glm::vec3 spot_direction;
	float spot_cutoff; // no spotlight if cutoff = 0
};

struct LightProps {
	Light light;
	glm::vec3 lightAmbient;
};

struct FogParams {
	glm::vec3 col; // Fog color
	float density; // For exp and exp2 equation
	float start; // This is only for linear fog
	float end; // This is only for linear fog
	int mode; // 1 = linear, 2 = exp, 3 = exp2
};

struct TextureProps {
	int useColorTexture;
};

int main(int argc, const char** argv) {
	const std::string applicationName = "Cube Mapping";
	
	vkcv::Core core = vkcv::Core::create(
			applicationName,
			VK_MAKE_VERSION(0, 0, 1),
			{ vk::QueueFlagBits::eGraphics, vk::QueueFlagBits::eTransfer },
			{ VK_KHR_SWAPCHAIN_EXTENSION_NAME }
	);
	
	vkcv::WindowHandle windowHandle = core.createWindow(applicationName, 800, 800, true);
	vkcv::Window& window = core.getWindow(windowHandle);
	
	vkcv::camera::CameraManager cameraManager (window);
	auto camHandle = cameraManager.addCamera(vkcv::camera::ControllerType::TRACKBALL);
	
	cameraManager.getCamera(camHandle).setNearFar(0.1f, 10.0f);
	
	vkcv::shader::GLSLCompiler compiler;
	vkcv::ShaderProgram cubeMappingShaderProgram, skyBoxShaderProgram;
	
	compiler.compileProgram(
			cubeMappingShaderProgram,
			{
					{ vkcv::ShaderStage::VERTEX, "shaders/CubeMapReflection.vert" },
					{ vkcv::ShaderStage::FRAGMENT, "shaders/CubeMapReflection.frag" }
			},
			nullptr
	);
	
	compiler.compileProgram(
			skyBoxShaderProgram,
			{
					{ vkcv::ShaderStage::VERTEX, "shaders/SkyBox.vert" },
					{ vkcv::ShaderStage::FRAGMENT, "shaders/SkyBox.frag" }
			},
			nullptr
	);
	
	vkcv::PassHandle renderPass = core.createPass(vkcv::PassConfig(
			{
					vkcv::AttachmentDescription(
							core.getSwapchainFormat(window.getSwapchain()),
							vkcv::AttachmentOperation::CLEAR,
							vkcv::AttachmentOperation::STORE,
							vk::ClearValue(vk::ClearColorValue(std::array<float, 4>{
									1.0f, 1.0f, 1.0f, 1.0f
							}))
					),
					vkcv::AttachmentDescription(
							vk::Format::eD32Sfloat,
							vkcv::AttachmentOperation::CLEAR,
							vkcv::AttachmentOperation::STORE
					)
			}
	));
	
	vkcv::PassHandle skyRenderPass = vkcv::passSwapchain(core, window.getSwapchain(), {
			vk::Format::eUndefined, vk::Format::eD32Sfloat
	}, false);
	
	const vkcv::VertexLayout vertexLayoutCubeMapping {
			vkcv::createVertexBindings(cubeMappingShaderProgram.getVertexAttachments())
	};
	
	const vkcv::VertexLayout vertexLayoutSykBox {
			vkcv::createVertexBindings(skyBoxShaderProgram.getVertexAttachments())
	};
	
	const vkcv::DescriptorBindings descriptorBindings = (
			cubeMappingShaderProgram.getReflectedDescriptors().at(0)
	);
	
	auto descriptorSetLayout = core.createDescriptorSetLayout(descriptorBindings);
	
	auto descriptorSetSphere = core.createDescriptorSet(descriptorSetLayout);
	auto descriptorSetTeapot = core.createDescriptorSet(descriptorSetLayout);
	
	const vkcv::DescriptorBindings& descriptorBindingsSkyBox = (
			skyBoxShaderProgram.getReflectedDescriptors().at(0)
	);
	
	auto descriptorSetLayoutSkyBox = core.createDescriptorSetLayout(descriptorBindingsSkyBox);
	auto descriptorSetSkyBox = core.createDescriptorSet(descriptorSetLayoutSkyBox);
	
	vkcv::GraphicsPipelineHandle cubeMappingPipeline = core.createGraphicsPipeline(
			vkcv::GraphicsPipelineConfig(
					cubeMappingShaderProgram,
					renderPass,
					vertexLayoutCubeMapping,
					{ descriptorSetLayout }
			)
	);
	
	vkcv::GraphicsPipelineHandle skyBoxPipeline = core.createGraphicsPipeline(
			vkcv::GraphicsPipelineConfig(
					skyBoxShaderProgram,
					skyRenderPass,
					vertexLayoutSykBox,
					{ descriptorSetLayoutSkyBox }
			)
	);
	
	auto matrixBuffer = vkcv::buffer<glm::mat4>(
			core,
			vkcv::BufferType::UNIFORM,
			2,
			vkcv::BufferMemoryType::HOST_VISIBLE
	);
	
	auto matBlueBuffer = vkcv::buffer<Material>(
			core,
			vkcv::BufferType::UNIFORM,
			1
	);
	
	Material mat_blue;
	mat_blue.diffuse = glm::vec3(0.0f, 0.0f, 1.0f);
	mat_blue.reflectionStrength = 0.7f;
	mat_blue.specular = glm::vec3(1.0f);
	mat_blue.shininess = 100.0f;
	matBlueBuffer.fill(&mat_blue, 1);
	
	auto matRedBuffer = vkcv::buffer<Material>(
			core,
			vkcv::BufferType::UNIFORM,
			1
	);
	
	Material mat_red;
	mat_red.diffuse = glm::vec3(1.0f, 0.0f, 0.0f);
	mat_red.reflectionStrength = 0.7f;
	mat_red.specular = glm::vec3(1.0f);
	mat_red.shininess = 100.0f;
	matRedBuffer.fill(&mat_red, 1);
	
	auto lightBuffer = vkcv::buffer<LightProps>(
			core,
			vkcv::BufferType::UNIFORM,
			1
	);
	
	LightProps lightProps;
	lightProps.light.pos = glm::vec4(0.0f, 1.0f, -10.0f, 1.0f);
	lightProps.light.col = glm::vec3(0.75f);
	lightProps.light.spot_exponent = 1.0f;
	lightProps.light.spot_direction = glm::vec3(0.0f);
	lightProps.light.spot_cutoff = 0.0f;
	lightProps.lightAmbient = glm::vec3(0.2f);
	lightBuffer.fill(&lightProps, 1);
	
	vkcv::asset::Texture negx = vkcv::asset::loadTexture("../../resources/cubeMap/negx.jpg");
	vkcv::asset::Texture negy = vkcv::asset::loadTexture("../../resources/cubeMap/negy.jpg");
	vkcv::asset::Texture negz = vkcv::asset::loadTexture("../../resources/cubeMap/negz.jpg");
	vkcv::asset::Texture posx = vkcv::asset::loadTexture("../../resources/cubeMap/posx.jpg");
	vkcv::asset::Texture posy = vkcv::asset::loadTexture("../../resources/cubeMap/posy.jpg");
	vkcv::asset::Texture posz = vkcv::asset::loadTexture("../../resources/cubeMap/posz.jpg");
	
	vkcv::ImageConfig cubemapConfig (negx.w, negx.h);
	cubemapConfig.setCubeMapImage(true);
	
	vkcv::Image cubemapImage = vkcv::image(core, vk::Format::eR8G8B8A8Unorm, cubemapConfig);
	cubemapImage.fillLayer(0, posx.data.data());
	cubemapImage.fillLayer(1, negx.data.data());
	cubemapImage.fillLayer(2, posy.data.data());
	cubemapImage.fillLayer(3, negy.data.data());
	cubemapImage.fillLayer(4, posz.data.data());
	cubemapImage.fillLayer(5, negz.data.data());
	
	vkcv::SamplerHandle sampler = vkcv::samplerLinear(core);
	
	{
		vkcv::DescriptorWrites writes;
		writes.writeUniformBuffer(0, matrixBuffer.getHandle());
		writes.writeUniformBuffer(1, matRedBuffer.getHandle());
		writes.writeUniformBuffer(2, lightBuffer.getHandle());
		writes.writeSampledImage(3, cubemapImage.getHandle());
		writes.writeSampler(4, sampler);
		core.writeDescriptorSet(descriptorSetSphere, writes);
	}
	
	{
		vkcv::DescriptorWrites writes;
		writes.writeUniformBuffer(0, matrixBuffer.getHandle());
		writes.writeUniformBuffer(1, matBlueBuffer.getHandle());
		writes.writeUniformBuffer(2, lightBuffer.getHandle());
		writes.writeSampledImage(3, cubemapImage.getHandle());
		writes.writeSampler(4, sampler);
		core.writeDescriptorSet(descriptorSetTeapot, writes);
	}
	
	{
		vkcv::DescriptorWrites writes;
		writes.writeUniformBuffer(0, matrixBuffer.getHandle());
		writes.writeSampledImage(1, cubemapImage.getHandle());
		writes.writeSampler(2, sampler);
		core.writeDescriptorSet(descriptorSetSkyBox, writes);
	}
	
	vkcv::gui::GUI gui (core, windowHandle);
	
	bool adjustedReflectionStrength = false;
	bool renderSkyBox = true;
	
	vkcv::geometry::Cuboid cube (glm::vec3(0.0f), 1.0f);
	vkcv::geometry::Sphere sphere (glm::vec3(0.0f), 0.3f);
	vkcv::geometry::Teapot teapot (glm::vec3(0.0f), 1.0f);
	
	sphere.setResolution(20);
	
	vkcv::InstanceDrawcall drawcallSphere (sphere.generateVertexData(core));
	drawcallSphere.useDescriptorSet(0, descriptorSetSphere);
	
	vkcv::InstanceDrawcall drawcallTeapot (teapot.generateVertexData(core));
	drawcallTeapot.useDescriptorSet(0, descriptorSetTeapot);
	
	vkcv::InstanceDrawcall drawcallSkyBox (cube.generateVertexData(core));
	drawcallSkyBox.useDescriptorSet(0, descriptorSetSkyBox);
	
	const vkcv::ImageHandle swapchainInput = vkcv::ImageHandle::createSwapchainImageHandle();
	
	vkcv::ImageHandle depthBuffer;
	
	glm::mat4* mvp = matrixBuffer.map();
	
	core.run([&](const vkcv::WindowHandle &windowHandle,
				 double t,
				 double dt,
				 uint32_t swapchainWidth,
				 uint32_t swapchainHeight
	) {
		if ((!depthBuffer) ||
			(swapchainWidth != core.getImageWidth(depthBuffer)) ||
			(swapchainHeight != core.getImageHeight(depthBuffer))) {
			depthBuffer = core.createImage(
					vk::Format::eD32Sfloat,
					vkcv::ImageConfig(
							swapchainWidth,
							swapchainHeight
					)
			);
		}
		
		if (adjustedReflectionStrength) {
			matBlueBuffer.fill(&mat_blue, 1);
			matRedBuffer.fill(&mat_red, 1);
			
			adjustedReflectionStrength = false;
		}
		
		cameraManager.update(dt);
		
		const glm::mat4 teapotMat = glm::rotate(
				glm::identity<glm::mat4>(),
				glm::radians(30.0f) * static_cast<float>(t),
				glm::vec3(0, 1, 0)
		);
		
		const glm::mat4 sphereMat = teapotMat * glm::rotate(
				glm::translate(glm::mat4(1.0f), glm::vec3(3, 0, 0)),
				glm::radians(90.0f) * static_cast<float>(t),
				glm::vec3(1, 1, 1)
		);
		
		mvp[0] = cameraManager.getActiveCamera().getView();
		mvp[1] = cameraManager.getActiveCamera().getProjection();
		
		auto pushConstants = vkcv::pushConstants<glm::mat4>();
		pushConstants.appendDrawcall(teapotMat);
		pushConstants.appendDrawcall(sphereMat);
		
		auto cmdStream = core.createCommandStream(vkcv::QueueType::Graphics);
		
		core.prepareImageForSampling(cmdStream, cubemapImage.getHandle());
		
		core.recordDrawcallsToCmdStream(
				cmdStream,
				cubeMappingPipeline,
				pushConstants,
				{
					drawcallTeapot,
					drawcallSphere
				},
				{
						swapchainInput,
						depthBuffer
				},
				windowHandle
		);
		
		if (renderSkyBox) {
			core.recordDrawcallsToCmdStream(
					cmdStream,
					skyBoxPipeline,
					vkcv::PushConstants(0),
					{ drawcallSkyBox },
					{
							swapchainInput,
							depthBuffer
					},
					windowHandle
			);
		}
		
		core.prepareSwapchainImageForPresent(cmdStream);
		core.submitCommandStream(cmdStream);
		
		gui.beginGUI();
		
		float reflectionStrength = mat_red.reflectionStrength;
		
		ImGui::Begin("Settings");
		ImGui::SliderFloat("Reflection strength", &reflectionStrength, 0.0f, 1.0f);
		ImGui::Checkbox("Render sky box", &renderSkyBox);
		ImGui::End();
		
		if (reflectionStrength != mat_red.reflectionStrength) {
			mat_blue.reflectionStrength = reflectionStrength;
			mat_red.reflectionStrength = reflectionStrength;
			
			adjustedReflectionStrength = true;
		}
		
		gui.endGUI();
	});
	
	matrixBuffer.unmap();
	return 0;
}
