
#include <random>
#include <vkcv/Buffer.hpp>
#include <vkcv/Core.hpp>
#include <vkcv/DescriptorWrites.hpp>
#include <vkcv/Window.hpp>
#include <vkcv/camera/CameraManager.hpp>
#include <vkcv/geometry/Teapot.hpp>
#include <vkcv/gui/GUI.hpp>
#include <vkcv/shader/GLSLCompiler.hpp>

struct InstanceParams {
	glm::vec3 offset;
	float pad0;
	glm::vec3 color;
	float pad1;
};

void generateInstances(int width, int height, int depth,
					   std::vector<InstanceParams>& instances) {
	std::default_random_engine randomEngine;
	instances.reserve(width * height * depth);
	
	size_t i, j, k;
	for (i = 0; i < width; i++) {
		for (j = 0; j < height; j++) {
			for (k = 0; k < depth; k++) {
				InstanceParams instance {};
				
				instance.offset = glm::vec3(
						(i * 3.0f) - width * 1.5f,
						(j * 2.0f) - height,
						(k * 2.0f) - depth
				);
				
				instance.color = glm::vec3(
						static_cast<float>(randomEngine() % 100 + 1) / 100.0f,
						static_cast<float>(randomEngine() % 100 + 1) / 100.0f,
						static_cast<float>(randomEngine() % 100 + 1) / 100.0f
				);
				
				instances.push_back(instance);
			}
		}
	}
}

int main(int argc, const char** argv) {
	const std::string applicationName = "Instancing Demo";
	
	vkcv::Core core = vkcv::Core::create(
			applicationName,
			VK_MAKE_VERSION(0, 0, 1),
			{ vk::QueueFlagBits::eGraphics, vk::QueueFlagBits::eTransfer },
			{ VK_KHR_SWAPCHAIN_EXTENSION_NAME }
	);
	
	vkcv::WindowHandle windowHandle = core.createWindow(applicationName, 800, 800, true);
	vkcv::Window& window = core.getWindow(windowHandle);
	
	vkcv::camera::CameraManager cameraManager (window);
	auto camHandle = cameraManager.addCamera(vkcv::camera::ControllerType::TRACKBALL);
	
	cameraManager.getCamera(camHandle).setNearFar(0.1f, 100.0f);
	
	vkcv::geometry::Teapot teapot (glm::vec3(0.0f), 1.0f);
	
	vkcv::shader::GLSLCompiler compiler;
	vkcv::ShaderProgram shaderProgram;
	
	compiler.compileProgram(
			shaderProgram,
			{
					{ vkcv::ShaderStage::VERTEX, "shaders/instancing.vert" },
					{ vkcv::ShaderStage::FRAGMENT, "shaders/instancing.frag" }
			},
			nullptr
	);
	
	vkcv::PassHandle renderPass = core.createPass(vkcv::PassConfig(
			{
					vkcv::AttachmentDescription(
							core.getSwapchainFormat(window.getSwapchain()),
							vkcv::AttachmentOperation::CLEAR,
							vkcv::AttachmentOperation::STORE,
							vk::ClearValue(vk::ClearColorValue(std::array<float, 4>{
									1.0f, 1.0f, 1.0f, 1.0f
							}))
					),
					vkcv::AttachmentDescription(
							vk::Format::eD32Sfloat,
							vkcv::AttachmentOperation::CLEAR,
							vkcv::AttachmentOperation::STORE
					)
			}
	));
	
	const vkcv::VertexLayout vertexLayout {
			vkcv::createVertexBindings(shaderProgram.getVertexAttachments())
	};
	
	const vkcv::DescriptorBindings descriptorBindings = (
			shaderProgram.getReflectedDescriptors().at(0)
	);
	
	auto descriptorSetLayout = core.createDescriptorSetLayout(descriptorBindings);
	auto descriptorSet = core.createDescriptorSet(descriptorSetLayout);
	
	vkcv::GraphicsPipelineHandle gfxPipeline = core.createGraphicsPipeline(
			vkcv::GraphicsPipelineConfig(
					shaderProgram,
					renderPass,
					vertexLayout,
					{ descriptorSetLayout }
			)
	);
	
	auto matrixBuffer = vkcv::buffer<glm::mat4>(
			core,
			vkcv::BufferType::UNIFORM,
			2,
			vkcv::BufferMemoryType::HOST_VISIBLE
	);
	
	int width = 5, height = 5, depth = 5;
	std::vector<InstanceParams> instances;
	
	vkcv::BufferHandle instanceBufferHandle;
	
	{
		vkcv::DescriptorWrites writes;
		writes.writeUniformBuffer(0, matrixBuffer.getHandle());
		core.writeDescriptorSet(descriptorSet, writes);
	}
	
	vkcv::gui::GUI gui (core, windowHandle);
	
	const vkcv::ImageHandle swapchainInput = vkcv::ImageHandle::createSwapchainImageHandle();
	
	vkcv::ImageHandle depthBuffer;
	
	glm::mat4* mvp = matrixBuffer.map();
	
	core.run([&](const vkcv::WindowHandle &windowHandle,
				 double t,
				 double dt,
				 uint32_t swapchainWidth,
				 uint32_t swapchainHeight
	) {
		if ((!depthBuffer) ||
			(swapchainWidth != core.getImageWidth(depthBuffer)) ||
			(swapchainHeight != core.getImageHeight(depthBuffer))) {
			depthBuffer = core.createImage(
					vk::Format::eD32Sfloat,
					vkcv::ImageConfig(
							swapchainWidth,
							swapchainHeight
					)
			);
		}
		
		cameraManager.update(dt);
		
		if (!instanceBufferHandle) {
			instances.clear();
			generateInstances(width, height, depth, instances);
			
			auto instanceBuffer = vkcv::buffer<InstanceParams>(
					core,
					vkcv::BufferType::STORAGE,
					instances.size()
			);
			
			instanceBuffer.fill(instances);
			instanceBufferHandle = instanceBuffer.getHandle();
			
			vkcv::DescriptorWrites writes;
			writes.writeStorageBuffer(1, instanceBufferHandle);
			core.writeDescriptorSet(descriptorSet, writes);
		}
		
		mvp[0] = cameraManager.getActiveCamera().getView();
		mvp[1] = cameraManager.getActiveCamera().getProjection();
		
		auto cmdStream = core.createCommandStream(vkcv::QueueType::Graphics);
		
		vkcv::InstanceDrawcall drawcall (teapot.generateVertexData(core), instances.size());
		drawcall.useDescriptorSet(0, descriptorSet);
		
		core.recordDrawcallsToCmdStream(
				cmdStream,
				gfxPipeline,
				vkcv::PushConstants(0),
				{ drawcall },
				{
						swapchainInput,
						depthBuffer
				},
				windowHandle
		);
		
		core.prepareSwapchainImageForPresent(cmdStream);
		core.submitCommandStream(cmdStream);
		
		gui.beginGUI();
		ImGui::Begin("Instances");
		
		int w = width, h = height, d = depth;
		
		ImGui::SliderInt("Width", &w, 1, 100);
		ImGui::SliderInt("Height", &h, 1, 100);
		ImGui::SliderInt("Depth", &d, 1, 100);
		
		ImGui::End();
		gui.endGUI();
		
		if ((w != width) || (h != height) || (d != depth)) {
			instanceBufferHandle = vkcv::BufferHandle();
			
			width = w;
			height = h;
			depth = d;
		}
	});
	
	matrixBuffer.unmap();
	return 0;
}