#version 450

layout(location = 0) in vec3 passPosition;
layout(location = 1) in vec3 passNormal;
layout(location = 2) in vec2 passTcoord;

layout(set=0, binding=0) uniform matrixBuffer {
	mat4 modelMatrix;
	mat4 viewMatrix;
	mat4 projectionMatrix;
};

struct Material {
	vec3 diffuse;
	float alpha;
	vec3 specular;
	float shininess;
};

layout(set=0, binding=1) uniform materialBuffer {
	Material mat;
};

struct Light {
	vec4 pos; // pos.w = 0 (dir light), pos.w = 1 (point light)
	vec3 col;
	float spot_exponent; 
	vec3 spot_direction; 
	float spot_cutoff; // no spotlight if cutoff = 0
};

layout(set=0, binding=2) uniform lightBuffer {
	Light light;
	vec3 lightAmbient;
};

struct FogParams {
	vec3 col; // Fog color
	float density; // For exp and exp2 equation
	float start; // This is only for linear fog
	float end; // This is only for linear fog
	int mode; // 1 = linear, 2 = exp, 3 = exp2
};

layout(set=0, binding=3) uniform fogBuffer {
	FogParams fog;
};

layout(set=0, binding=4) uniform texturePropsBuffer {
	int useColorTexture;
};

layout(set=0, binding=5) uniform texture2D	colorTexture;
layout(set=0, binding=6) uniform sampler	colorSampler;

layout(location = 0) out vec4 fragmentColor;

float getFogFactor(FogParams fog, float z) { 
	float f = 0.0; 
	if (fog.mode == 1) { 
		f = (fog.end-z)/(fog.end-fog.start); 
	} else if (fog.mode == 2) {
		f = exp(-fog.density*z); 
	} else if(fog.mode == 3) {
		f = exp(-pow(fog.density*z, 2.0));
	}
	
	f = clamp(f, 0.0, 1.0); 
	
	return f; 
}

void main() { 
	vec3 lightVector;
	float spot;
	vec3 diffuse_color;
	float diffuse_alpha;

	// Diffuse
	vec3 light_camcoord = (light.pos).xyz;
	if (light.pos.w > 0.001f) {
		lightVector = normalize(light_camcoord - passPosition);
	} else {
		lightVector = normalize(light_camcoord);
	}
	
	float cos_phi = max(dot(passNormal,lightVector), 0.0f);

	// Specular
	vec3 eye = normalize(-passPosition);
	vec3 reflection = normalize(reflect(-lightVector, passNormal));
	float cos_psi_n = pow(max(dot(reflection, eye), 0.0f), mat.shininess);

	// Spotlight	
	if (light.spot_cutoff < 0.001f) {
		spot = 1.0;
	} else {
		float cos_phi_spot = max(dot(-lightVector, mat3(viewMatrix) * light.spot_direction), 0.0f);
		if (cos_phi_spot >= cos(light.spot_cutoff)) {
			spot = pow(cos_phi_spot, light.spot_exponent);
		} else {
			spot = 0.0f;
		}
	}

	// Textures 
	if (useColorTexture != 0) {
		diffuse_color = texture(sampler2D(colorTexture, colorSampler), passTcoord).rgb;
		diffuse_alpha = texture(sampler2D(colorTexture, colorSampler), passTcoord).a;
	} else {
		diffuse_color = mat.diffuse;
		diffuse_alpha = mat.alpha;
	}

	// All together 
	fragmentColor.rgb = diffuse_color * lightAmbient;
	fragmentColor.rgb += spot * diffuse_color * cos_phi * light.col;
	fragmentColor.rgb += spot * mat.specular * cos_psi_n * light.col;
	fragmentColor.a = diffuse_alpha;

	//Add fog
	if (fog.mode != 0) {
		float f = getFogFactor(fog, length(passPosition));
		fragmentColor.rgb = f * fragmentColor.rgb + (1-f) * fog.col;
	}
}
