
#include <iostream>
#include <vkcv/Buffer.hpp>
#include <vkcv/Core.hpp>
#include <vkcv/DescriptorWrites.hpp>
#include <vkcv/Image.hpp>
#include <vkcv/Sampler.hpp>
#include <vkcv/Window.hpp>
#include <vkcv/asset/asset_loader.hpp>
#include <vkcv/camera/CameraManager.hpp>
#include <vkcv/geometry/Cuboid.hpp>
#include <vkcv/gui/GUI.hpp>
#include <vkcv/shader/GLSLCompiler.hpp>

struct Material {
	glm::vec3 diffuse;
	float alpha;
	glm::vec3 specular;
	float shininess;
};

struct Light {
	glm::vec4 pos; // pos.w = 0 (dir light), pos.w = 1 (point light)
	glm::vec3 col;
	float spot_exponent;
	glm::vec3 spot_direction;
	float spot_cutoff; // no spotlight if cutoff = 0
};

struct LightProps {
	Light light;
	glm::vec3 lightAmbient;
};

struct FogParams {
	glm::vec3 col; // Fog color
	float density; // For exp and exp2 equation
	float start; // This is only for linear fog
	float end; // This is only for linear fog
	int mode; // 1 = linear, 2 = exp, 3 = exp2
};

struct TextureProps {
	int useColorTexture;
};

int main(int argc, const char** argv) {
	const std::string applicationName = "Normal Mapping";
	
	vkcv::Core core = vkcv::Core::create(
			applicationName,
			VK_MAKE_VERSION(0, 0, 1),
			{ vk::QueueFlagBits::eGraphics, vk::QueueFlagBits::eTransfer },
			{ VK_KHR_SWAPCHAIN_EXTENSION_NAME }
	);
	
	vkcv::WindowHandle windowHandle = core.createWindow(applicationName, 800, 800, true);
	vkcv::Window& window = core.getWindow(windowHandle);
	
	vkcv::camera::CameraManager cameraManager (window);
	auto camHandle = cameraManager.addCamera(vkcv::camera::ControllerType::TRACKBALL);
	
	vkcv::geometry::Cuboid cube (glm::vec3(0.0f), 1.0f);
	
	vkcv::shader::GLSLCompiler compiler;
	vkcv::ShaderProgram phongShaderProgram, normalMappingShaderProgram;
	
	compiler.compileProgram(
			phongShaderProgram,
			{
				{ vkcv::ShaderStage::VERTEX, "shaders/Phong.vert" },
				{ vkcv::ShaderStage::FRAGMENT, "shaders/Phong.frag" }
			},
			nullptr
	);
	
	compiler.compileProgram(
			normalMappingShaderProgram,
			{
					{ vkcv::ShaderStage::VERTEX, "shaders/NormalMapping.vert" },
					{ vkcv::ShaderStage::FRAGMENT, "shaders/NormalMapping.frag" }
			},
			nullptr
	);
	
	vkcv::PassHandle renderPass = core.createPass(vkcv::PassConfig(
			{
				vkcv::AttachmentDescription(
						core.getSwapchainFormat(window.getSwapchain()),
						vkcv::AttachmentOperation::CLEAR,
						vkcv::AttachmentOperation::STORE,
						vk::ClearValue(vk::ClearColorValue(std::array<float, 4>{
							1.0f, 1.0f, 1.0f, 1.0f
						}))
				),
				vkcv::AttachmentDescription(
						vk::Format::eD32Sfloat,
						vkcv::AttachmentOperation::CLEAR,
						vkcv::AttachmentOperation::STORE
				)
			}
	));
	
	const vkcv::VertexLayout vertexLayoutPhong {
			vkcv::createVertexBindings(phongShaderProgram.getVertexAttachments())
	};
	
	const vkcv::VertexLayout vertexLayoutNormalMapping {
			vkcv::createVertexBindings(normalMappingShaderProgram.getVertexAttachments())
	};
	
	const vkcv::DescriptorBindings descriptorBindings0 = (
			phongShaderProgram.getReflectedDescriptors().at(0)
	);
	
	auto descriptorSetLayout0 = core.createDescriptorSetLayout(descriptorBindings0);
	auto descriptorSet0 = core.createDescriptorSet(descriptorSetLayout0);
	
	const vkcv::DescriptorBindings& descriptorBindings1 = (
			normalMappingShaderProgram.getReflectedDescriptors().at(1)
	);
	
	auto descriptorSetLayout1 = core.createDescriptorSetLayout(descriptorBindings1);
	auto descriptorSet1 = core.createDescriptorSet(descriptorSetLayout1);
	
	vkcv::GraphicsPipelineHandle phongPipeline = core.createGraphicsPipeline(
			vkcv::GraphicsPipelineConfig(
					phongShaderProgram,
					renderPass,
					vertexLayoutPhong,
					{ descriptorSetLayout0 }
			)
	);
	
	vkcv::GraphicsPipelineHandle normalMappingPipeline = core.createGraphicsPipeline(
			vkcv::GraphicsPipelineConfig(
					normalMappingShaderProgram,
					renderPass,
					vertexLayoutNormalMapping,
					{
						descriptorSetLayout0,
						descriptorSetLayout1
					}
			)
	);
	
	auto matrixBuffer = vkcv::buffer<glm::mat4>(
			core,
			vkcv::BufferType::UNIFORM,
			3,
			vkcv::BufferMemoryType::HOST_VISIBLE
	);
	
	auto materialBuffer = vkcv::buffer<Material>(
			core,
			vkcv::BufferType::UNIFORM,
			1
	);
	
	Material material;
	material.diffuse = glm::vec3(1.0f, 0.0f, 0.0f);
	material.alpha = 1.0f;
	material.specular = glm::vec3(1.0f);
	material.shininess = 120.0f;
	materialBuffer.fill(&material, 1);
	
	auto lightBuffer = vkcv::buffer<LightProps>(
			core,
			vkcv::BufferType::UNIFORM,
			1
	);
	
	LightProps lightProps;
	lightProps.light.pos = glm::vec4(0.0f, 1.0f, -10.0f, 1.0f);
	lightProps.light.col = glm::vec3(0.75f);
	lightProps.light.spot_exponent = 1.0f;
	lightProps.light.spot_direction = glm::vec3(0.0f);
	lightProps.light.spot_cutoff = 0.0f;
	lightProps.lightAmbient = glm::vec3(0.2f);
	lightBuffer.fill(&lightProps, 1);
	
	auto fogBuffer = vkcv::buffer<FogParams>(
			core,
			vkcv::BufferType::UNIFORM,
			1
	);
	
	FogParams fogParams;
	fogParams.col = glm::vec3(1.0f);
	fogParams.density = 1.0f;
	fogParams.start = 1.0f;
	fogParams.end = 10.0f;
	fogParams.mode = 0;
	fogBuffer.fill(&fogParams, 1);
	
	auto texturePropsBuffer = vkcv::buffer<TextureProps>(
			core,
			vkcv::BufferType::UNIFORM,
			1,
			vkcv::BufferMemoryType::HOST_VISIBLE
	);
	
	vkcv::asset::Texture diffusemap = vkcv::asset::loadTexture("../../resources/normalmapping/diffusemap.png");
	vkcv::asset::Texture normalmap = vkcv::asset::loadTexture("../../resources/normalmapping/normalmap.png");
	
	vkcv::Image diffusemapImage = vkcv::image(core, vk::Format::eR8G8B8A8Unorm, diffusemap.w, diffusemap.h);
	diffusemapImage.fill(diffusemap.data.data());
	
	vkcv::Image normalmapImage = vkcv::image(core, vk::Format::eR8G8B8A8Unorm, normalmap.w, normalmap.h);
	normalmapImage.fill(normalmap.data.data());
	
	vkcv::SamplerHandle sampler = vkcv::samplerLinear(core);
	
	{
		vkcv::DescriptorWrites writes;
		writes.writeUniformBuffer(0, matrixBuffer.getHandle());
		writes.writeUniformBuffer(1, materialBuffer.getHandle());
		writes.writeUniformBuffer(2, lightBuffer.getHandle());
		writes.writeUniformBuffer(3, fogBuffer.getHandle());
		writes.writeUniformBuffer(4, texturePropsBuffer.getHandle());
		writes.writeSampledImage(5, diffusemapImage.getHandle());
		writes.writeSampler(6, sampler);
		core.writeDescriptorSet(descriptorSet0, writes);
	}
	
	{
		vkcv::DescriptorWrites writes;
		writes.writeSampledImage(0, normalmapImage.getHandle());
		core.writeDescriptorSet(descriptorSet1, writes);
	}
	
	vkcv::gui::GUI gui (core, windowHandle);
	
	bool useColorTexture = false;
	bool useNormalMappingShader = false;
	
	const auto vertexData = cube.generateVertexData(core);
	
	vkcv::InstanceDrawcall drawcallPhong (vertexData);
	drawcallPhong.useDescriptorSet(0, descriptorSet0);
	
	vkcv::InstanceDrawcall drawcallNormalMapping (vertexData);
	drawcallNormalMapping.useDescriptorSet(0, descriptorSet0);
	drawcallNormalMapping.useDescriptorSet(1, descriptorSet1);
	
	const vkcv::ImageHandle swapchainInput = vkcv::ImageHandle::createSwapchainImageHandle();
	
	vkcv::ImageHandle depthBuffer;
	
	glm::mat4* mvp = matrixBuffer.map();
	TextureProps* textureProps = texturePropsBuffer.map();
	
	core.run([&](const vkcv::WindowHandle &windowHandle,
				 double t,
				 double dt,
				 uint32_t swapchainWidth,
				 uint32_t swapchainHeight
	) {
		if ((!depthBuffer) ||
			(swapchainWidth != core.getImageWidth(depthBuffer)) ||
			(swapchainHeight != core.getImageHeight(depthBuffer))) {
			depthBuffer = core.createImage(
					vk::Format::eD32Sfloat,
					vkcv::ImageConfig(
							swapchainWidth,
							swapchainHeight
					)
			);
		}
		
		cameraManager.update(dt);
		
		mvp[0] = glm::rotate(glm::identity<glm::mat4>(), glm::pi<float>() / 2, glm::vec3(1, 0, 0));
		mvp[1] = cameraManager.getActiveCamera().getView();
		mvp[2] = cameraManager.getActiveCamera().getProjection();
		
		textureProps->useColorTexture = useColorTexture? 1 : 0;
		
		auto cmdStream = core.createCommandStream(vkcv::QueueType::Graphics);
		
		core.prepareImageForSampling(cmdStream, diffusemapImage.getHandle());
		
		if (useNormalMappingShader) {
			core.prepareImageForSampling(cmdStream, normalmapImage.getHandle());
		}
		
		core.recordDrawcallsToCmdStream(
				cmdStream,
				useNormalMappingShader? normalMappingPipeline : phongPipeline,
				vkcv::PushConstants(0),
				{
					useNormalMappingShader? drawcallNormalMapping : drawcallPhong
				},
				{
					swapchainInput,
					depthBuffer
				},
				windowHandle
		);
		
		core.prepareSwapchainImageForPresent(cmdStream);
		core.submitCommandStream(cmdStream);
		
		gui.beginGUI();
		ImGui::Begin("Settings");
		ImGui::Checkbox("Use color texture", &useColorTexture);
		ImGui::Checkbox("Use normal mapping shader", &useNormalMappingShader);
		ImGui::End();
		gui.endGUI();
	});
	
	matrixBuffer.unmap();
	texturePropsBuffer.unmap();
	return 0;
}