
#include <vkcv/Buffer.hpp>
#include <vkcv/Core.hpp>
#include <vkcv/DescriptorWrites.hpp>
#include <vkcv/Image.hpp>
#include <vkcv/Pass.hpp>
#include <vkcv/Sampler.hpp>
#include <vkcv/Window.hpp>
#include <vkcv/asset/asset_loader.hpp>
#include <vkcv/camera/CameraManager.hpp>
#include <vkcv/geometry/Cuboid.hpp>
#include <vkcv/geometry/Sphere.hpp>
#include <vkcv/gui/GUI.hpp>
#include <vkcv/shader/GLSLCompiler.hpp>

struct Material {
	glm::vec3 diffuse;
	float alpha;
};

struct TextureProps {
	int useColorTexture;
};

int main(int argc, const char** argv) {
	const std::string applicationName = "SSAO";
	
	vkcv::Features features;
	features.requireExtension(VK_KHR_SWAPCHAIN_EXTENSION_NAME);
	features.requireExtensionFeature<vk::PhysicalDeviceShaderTerminateInvocationFeatures>(
			VK_KHR_SHADER_TERMINATE_INVOCATION_EXTENSION_NAME,
			[](vk::PhysicalDeviceShaderTerminateInvocationFeatures& features) {
				features.setShaderTerminateInvocation(true);
			}
	);
	
	vkcv::Core core = vkcv::Core::create(
			applicationName,
			VK_MAKE_VERSION(0, 0, 1),
			{ vk::QueueFlagBits::eGraphics, vk::QueueFlagBits::eTransfer },
			features
	);
	
	vkcv::WindowHandle windowHandle = core.createWindow(applicationName, 800, 800, true);
	vkcv::Window& window = core.getWindow(windowHandle);
	
	vkcv::camera::CameraManager cameraManager (window);
	auto camHandle = cameraManager.addCamera(vkcv::camera::ControllerType::TRACKBALL);
	
	cameraManager.getCamera(camHandle).setNearFar(0.1f, 100.0f);
	
	vkcv::geometry::Cuboid plane (glm::vec3(0.0f), glm::vec3(1.0f, 0.01f, 1.0f));
	
	vkcv::geometry::Sphere sphere (glm::vec3(0.0f), 1.0f);
	sphere.setResolution(20);
	
	vkcv::shader::GLSLCompiler compiler;
	vkcv::ShaderProgram shaderProgram, lightProgram, aoProgram, blurProgram, compositeProgram;
	
	compiler.compileProgram(
			shaderProgram,
			{
					{ vkcv::ShaderStage::VERTEX, "shaders/GBuffer.vert" },
					{ vkcv::ShaderStage::FRAGMENT, "shaders/GBuffer.frag" }
			},
			nullptr
	);
	
	compiler.compileProgram(
			lightProgram,
			{
					{ vkcv::ShaderStage::VERTEX, "shaders/lighting.vert" },
					{ vkcv::ShaderStage::FRAGMENT, "shaders/lighting.frag" }
			},
			nullptr
	);
	
	compiler.compileProgram(
			aoProgram,
			{
					{ vkcv::ShaderStage::VERTEX, "shaders/screenFill.vert" },
					{ vkcv::ShaderStage::FRAGMENT, "shaders/ambientOcclusion.frag" }
			},
			nullptr
	);
	
	compiler.compileProgram(
			blurProgram,
			{
					{ vkcv::ShaderStage::VERTEX, "shaders/screenFill.vert" },
					{ vkcv::ShaderStage::FRAGMENT, "shaders/blur1D.frag" }
			},
			nullptr
	);
	
	compiler.compileProgram(
			compositeProgram,
			{
					{ vkcv::ShaderStage::VERTEX, "shaders/screenFill.vert" },
					{ vkcv::ShaderStage::FRAGMENT, "shaders/finalCompositing.frag" }
			},
			nullptr
	);
	
	const vkcv::PassConfig gPassConfig ({
		vkcv::AttachmentDescription(
				vk::Format::eR32G32B32A32Sfloat,
				vkcv::AttachmentOperation::CLEAR,
				vkcv::AttachmentOperation::STORE,
				vk::ClearValue(vk::ClearColorValue(std::array<float, 4>{
						0.0f, 0.0f, 0.0f, 0.0f
				}))
		),
		vkcv::AttachmentDescription(
				vk::Format::eR32G32B32A32Sfloat,
				vkcv::AttachmentOperation::CLEAR,
				vkcv::AttachmentOperation::STORE,
				vk::ClearValue(vk::ClearColorValue(std::array<float, 4>{
						0.5f, 0.5f, 0.5f, 0.5f
				}))
		),
		vkcv::AttachmentDescription(
				core.getSwapchainFormat(window.getSwapchain()),
				vkcv::AttachmentOperation::CLEAR,
				vkcv::AttachmentOperation::STORE,
				vk::ClearValue(vk::ClearColorValue(std::array<float, 4>{
						1.0f, 1.0f, 1.0f, 1.0f
				}))
		),
		vkcv::AttachmentDescription(
				vk::Format::eD32Sfloat,
				vkcv::AttachmentOperation::CLEAR,
				vkcv::AttachmentOperation::STORE
		)
	});
	
	vkcv::PassHandle gBufferPass = core.createPass(gPassConfig);
	
	vkcv::PassHandle gBufferPositionPass = core.createPass(vkcv::PassConfig(
			{
					vkcv::AttachmentDescription(
							core.getSwapchainFormat(window.getSwapchain()),
							vkcv::AttachmentOperation::CLEAR,
							vkcv::AttachmentOperation::STORE,
							vk::ClearValue(vk::ClearColorValue(std::array<float, 4>{
									0.0f, 0.0f, 0.0f, 0.0f
							}))
					),
					gPassConfig.getAttachments()[1],
					gPassConfig.getAttachments()[2],
					gPassConfig.getAttachments()[3]
			}
	));
	
	vkcv::PassHandle gBufferNormalPass = core.createPass(vkcv::PassConfig(
			{
					gPassConfig.getAttachments()[0],
					vkcv::AttachmentDescription(
							core.getSwapchainFormat(window.getSwapchain()),
							vkcv::AttachmentOperation::CLEAR,
							vkcv::AttachmentOperation::STORE,
							vk::ClearValue(vk::ClearColorValue(std::array<float, 4>{
									0.5f, 0.5f, 0.5f, 0.5f
							}))
					),
					gPassConfig.getAttachments()[2],
					gPassConfig.getAttachments()[3]
			}
	));
	
	vkcv::PassHandle simplePass = vkcv::passSwapchain(core, window.getSwapchain(), {
		vk::Format::eUndefined
	});
	
	const vkcv::VertexLayout vertexLayout {
			vkcv::createVertexBindings(shaderProgram.getVertexAttachments())
	};
	
	const vkcv::DescriptorBindings descriptorBindings = (
			shaderProgram.getReflectedDescriptors().at(0)
	);
	
	auto descriptorSetLayout = core.createDescriptorSetLayout(descriptorBindings);
	auto descriptorSet = core.createDescriptorSet(descriptorSetLayout);
	
	vkcv::GraphicsPipelineHandle gBufferPipeline = core.createGraphicsPipeline(
			vkcv::GraphicsPipelineConfig(
					shaderProgram,
					gBufferPass,
					vertexLayout,
					{ descriptorSetLayout }
			)
	);
	
	vkcv::GraphicsPipelineHandle gBufferPosPipeline = core.createGraphicsPipeline(
			vkcv::GraphicsPipelineConfig(
					shaderProgram,
					gBufferPositionPass,
					vertexLayout,
					{ descriptorSetLayout }
			)
	);
	
	vkcv::GraphicsPipelineHandle gBufferNorPipeline = core.createGraphicsPipeline(
			vkcv::GraphicsPipelineConfig(
					shaderProgram,
					gBufferNormalPass,
					vertexLayout,
					{ descriptorSetLayout }
			)
	);
	
	const vkcv::VertexLayout lightVertexLayout {
			vkcv::createVertexBindings(lightProgram.getVertexAttachments())
	};
	
	const vkcv::DescriptorBindings lightDescSetBindings = (
			lightProgram.getReflectedDescriptors().at(0)
	);
	
	auto lightDescSetLayout = core.createDescriptorSetLayout(lightDescSetBindings);
	auto lightDescSet = core.createDescriptorSet(lightDescSetLayout);
	
	vkcv::GraphicsPipelineConfig lightPipelineConfig (
			lightProgram,
			simplePass,
			lightVertexLayout,
			{ lightDescSetLayout }
	);
	
	lightPipelineConfig.setBlendMode(vkcv::BlendMode::Additive);
	
	vkcv::GraphicsPipelineHandle lightPipeline = core.createGraphicsPipeline(
			lightPipelineConfig
	);
	
	const vkcv::DescriptorBindings aoDescSetBindings = (
			aoProgram.getReflectedDescriptors().at(0)
	);
	
	auto aoDescSetLayout = core.createDescriptorSetLayout(aoDescSetBindings);
	auto aoDescSet = core.createDescriptorSet(aoDescSetLayout);
	
	vkcv::GraphicsPipelineHandle aoPipeline = core.createGraphicsPipeline(
			vkcv::GraphicsPipelineConfig(
					aoProgram,
					simplePass,
					{},
					{ aoDescSetLayout }
			)
	);
	
	const vkcv::DescriptorBindings blurDescSetBindings = (
			blurProgram.getReflectedDescriptors().at(0)
	);
	
	auto blurDescSetLayout = core.createDescriptorSetLayout(blurDescSetBindings);
	auto blurDescSet = core.createDescriptorSet(blurDescSetLayout);
	
	vkcv::GraphicsPipelineHandle blurPipeline = core.createGraphicsPipeline(
			vkcv::GraphicsPipelineConfig(
					blurProgram,
					simplePass,
					{},
					{ blurDescSetLayout }
			)
	);
	
	const vkcv::DescriptorBindings compositeDescSetBindings = (
			compositeProgram.getReflectedDescriptors().at(0)
	);
	
	auto compositeDescSetLayout = core.createDescriptorSetLayout(compositeDescSetBindings);
	auto compositeDescSet = core.createDescriptorSet(compositeDescSetLayout);
	
	vkcv::GraphicsPipelineHandle compositePipeline = core.createGraphicsPipeline(
			vkcv::GraphicsPipelineConfig(
					compositeProgram,
					simplePass,
					{},
					{ compositeDescSetLayout }
			)
	);
	
	auto matrixBuffer = vkcv::buffer<glm::mat4>(
			core,
			vkcv::BufferType::UNIFORM,
			2,
			vkcv::BufferMemoryType::HOST_VISIBLE
	);
	
	auto materialBuffer = vkcv::buffer<Material>(
			core,
			vkcv::BufferType::UNIFORM,
			1
	);
	
	Material material;
	material.diffuse = glm::vec3(1.0f, 0.0f, 0.0f);
	material.alpha = 1.0f;
	materialBuffer.fill(&material, 1);
	
	auto texturePropsBuffer = vkcv::buffer<TextureProps>(
			core,
			vkcv::BufferType::UNIFORM,
			1,
			vkcv::BufferMemoryType::HOST_VISIBLE
	);
	
	vkcv::asset::Texture cv_logo = vkcv::asset::loadTexture("../../resources/cv_logo.png");
	
	vkcv::Image cv_logoImage = vkcv::image(core, vk::Format::eR8G8B8A8Unorm, cv_logo.w, cv_logo.h);
	cv_logoImage.fill(cv_logo.data.data());
	
	vkcv::SamplerHandle sampler = vkcv::samplerLinear(core);
	
	auto resolutionBuffer = vkcv::buffer<glm::vec2>(
			core,
			vkcv::BufferType::UNIFORM,
			1,
			vkcv::BufferMemoryType::HOST_VISIBLE
	);
	
	int sizeX = 7, sizeZ = 7;
	vkcv::PushConstants pushConstants = vkcv::pushConstants<glm::mat4>();
	
	{
		vkcv::DescriptorWrites writes;
		writes.writeUniformBuffer(0, matrixBuffer.getHandle());
		writes.writeUniformBuffer(1, materialBuffer.getHandle());
		writes.writeUniformBuffer(2, texturePropsBuffer.getHandle());
		writes.writeSampledImage(3, cv_logoImage.getHandle());
		writes.writeSampler(4, sampler);
		core.writeDescriptorSet(descriptorSet, writes);
	}
	
	{
		vkcv::DescriptorWrites writes;
		writes.writeUniformBuffer(0, matrixBuffer.getHandle());
		writes.writeSampler(3, sampler);
		core.writeDescriptorSet(aoDescSet, writes);
		
		writes.writeUniformBuffer(4, resolutionBuffer.getHandle());
		core.writeDescriptorSet(lightDescSet, writes);
	}
	
	{
		vkcv::DescriptorWrites writes;
		writes.writeSampler(1, sampler);
		core.writeDescriptorSet(blurDescSet, writes);
	}
	
	{
		vkcv::DescriptorWrites writes;
		writes.writeSampler(3, sampler);
		core.writeDescriptorSet(compositeDescSet, writes);
	}
	
	vkcv::gui::GUI gui (core, windowHandle);
	
	bool useColorTexture = true;
	int showBuffer = 0;
	bool useBlurPass = false;
	
	const std::vector<const char*> showBufferNames = {
			"Composition",
			"Position",
			"Normal",
			"Color",
			"Light",
			"Ambient Occlusion"
	};
	
	float ssaoRadius = 0.2f;
	float ssaoQuality = 16.0f;
	
	bool useLighting = true;
	bool useSSAO = true;
	
	vkcv::InstanceDrawcall planeDrawcall (plane.generateVertexData(core));
	planeDrawcall.useDescriptorSet(0, descriptorSet);
	
	vkcv::InstanceDrawcall sphereDrawcall (sphere.generateVertexData(core));
	sphereDrawcall.useDescriptorSet(0, descriptorSet);
	
	std::vector<vkcv::InstanceDrawcall> lightDrawcalls;
	lightDrawcalls.reserve(100);
	
	sphere.setRadius(3.0f);
	vkcv::VertexData lightSphereData = sphere.generateVertexData(core);
	
	for (size_t i = 0; i < 100; i++) {
		vkcv::InstanceDrawcall lightDrawcall (lightSphereData);
		lightDrawcall.useDescriptorSet(0, lightDescSet);
		lightDrawcalls.push_back(lightDrawcall);
	}
	
	vkcv::VertexData simpleVertexData;
	simpleVertexData.setCount(3);
	
	vkcv::InstanceDrawcall aoDrawcall (simpleVertexData);
	aoDrawcall.useDescriptorSet(0, aoDescSet);
	
	vkcv::InstanceDrawcall blurDrawcall (simpleVertexData);
	blurDrawcall.useDescriptorSet(0, blurDescSet);
	
	vkcv::InstanceDrawcall compositeDrawcall (simpleVertexData);
	compositeDrawcall.useDescriptorSet(0, compositeDescSet);
	
	const vkcv::ImageHandle swapchainInput = vkcv::ImageHandle::createSwapchainImageHandle();
	
	vkcv::ImageHandle positionBuffer;
	vkcv::ImageHandle normalBuffer;
	vkcv::ImageHandle colorBuffer;
	vkcv::ImageHandle depthBuffer;
	
	vkcv::ImageHandle lightBuffer;
	vkcv::ImageHandle aoBuffer;
	vkcv::ImageHandle blurBuffer;
	
	glm::mat4* mvp = matrixBuffer.map();
	TextureProps* textureProps = texturePropsBuffer.map();
	glm::vec2* res = resolutionBuffer.map();
	
	core.run([&](const vkcv::WindowHandle &windowHandle,
				 double t,
				 double dt,
				 uint32_t swapchainWidth,
				 uint32_t swapchainHeight
	) {
		vkcv::ImageConfig bufferConfig (
				swapchainWidth,
				swapchainHeight
		);
		
		bufferConfig.setSupportingColorAttachment(true);
		
		if ((!positionBuffer) ||
			(swapchainWidth != core.getImageWidth(positionBuffer)) ||
			(swapchainHeight != core.getImageHeight(positionBuffer))) {
			positionBuffer = core.createImage(
					vk::Format::eR32G32B32A32Sfloat,
					bufferConfig
			);
			
			vkcv::DescriptorWrites writes;
			writes.writeSampledImage(1, positionBuffer);
			core.writeDescriptorSet(lightDescSet, writes);
			core.writeDescriptorSet(aoDescSet, writes);
		}
		
		if ((!normalBuffer) ||
			(swapchainWidth != core.getImageWidth(normalBuffer)) ||
			(swapchainHeight != core.getImageHeight(normalBuffer))) {
			normalBuffer = core.createImage(
					vk::Format::eR32G32B32A32Sfloat,
					bufferConfig
			);
			
			vkcv::DescriptorWrites writes;
			writes.writeSampledImage(2, normalBuffer);
			core.writeDescriptorSet(lightDescSet, writes);
			core.writeDescriptorSet(aoDescSet, writes);
		}
		
		if ((!colorBuffer) ||
			(swapchainWidth != core.getImageWidth(colorBuffer)) ||
			(swapchainHeight != core.getImageHeight(colorBuffer))) {
			colorBuffer = core.createImage(
					core.getSwapchainFormat(core.getWindow(windowHandle).getSwapchain()),
					bufferConfig
			);
			
			vkcv::DescriptorWrites writes;
			writes.writeSampledImage(1, colorBuffer);
			core.writeDescriptorSet(compositeDescSet, writes);
		}
		
		if ((!depthBuffer) ||
			(swapchainWidth != core.getImageWidth(depthBuffer)) ||
			(swapchainHeight != core.getImageHeight(depthBuffer))) {
			depthBuffer = core.createImage(
					vk::Format::eD32Sfloat,
					vkcv::ImageConfig(
							swapchainWidth,
							swapchainHeight
					)
			);
		}
		
		if ((!lightBuffer) ||
			(swapchainWidth != core.getImageWidth(lightBuffer)) ||
			(swapchainHeight != core.getImageHeight(lightBuffer))) {
			lightBuffer = core.createImage(
					core.getSwapchainFormat(core.getWindow(windowHandle).getSwapchain()),
					bufferConfig
			);
			
			vkcv::DescriptorWrites writes;
			writes.writeSampledImage(0, lightBuffer);
			core.writeDescriptorSet(compositeDescSet, writes);
		}
		
		if ((!aoBuffer) ||
			(swapchainWidth != core.getImageWidth(aoBuffer)) ||
			(swapchainHeight != core.getImageHeight(aoBuffer))) {
			aoBuffer = core.createImage(
					core.getSwapchainFormat(core.getWindow(windowHandle).getSwapchain()),
					bufferConfig
			);
			
			vkcv::DescriptorWrites writes;
			writes.writeSampledImage(2, aoBuffer);
			core.writeDescriptorSet(compositeDescSet, writes);
		}
		
		if ((!blurBuffer) ||
			(swapchainWidth != core.getImageWidth(blurBuffer)) ||
			(swapchainHeight != core.getImageHeight(blurBuffer))) {
			blurBuffer = core.createImage(
					core.getSwapchainFormat(core.getWindow(windowHandle).getSwapchain()),
					bufferConfig
			);
			
			vkcv::DescriptorWrites writes;
			writes.writeSampledImage(0, blurBuffer);
			core.writeDescriptorSet(blurDescSet, writes);
		}
		
		cameraManager.update(dt);
		
		mvp[0] = cameraManager.getActiveCamera().getView();
		mvp[1] = cameraManager.getActiveCamera().getProjection();
		
		textureProps->useColorTexture = useColorTexture? 1 : 0;
		
		res[0] = glm::vec2(swapchainWidth, swapchainHeight);
		
		auto cmdStream = core.createCommandStream(vkcv::QueueType::Graphics);
		
		std::vector<vkcv::InstanceDrawcall> drawcalls;
		drawcalls.reserve(sizeX * sizeZ + 1);
		drawcalls.push_back(planeDrawcall);
		
		pushConstants.clear();
		pushConstants.appendDrawcall(glm::scale(glm::translate(
				glm::identity<glm::mat4>(),
				glm::vec3(0, -2.5f, 0)
		), glm::vec3(sizeX + 4, 1.0f, sizeZ + 4)));
		
		for (size_t i = 0; i < sizeX; i++) {
			for (size_t j = 0; j < sizeZ; j++) {
				drawcalls.push_back(sphereDrawcall);
				
				pushConstants.appendDrawcall(glm::scale(glm::translate(
						glm::identity<glm::mat4>(),
						glm::vec3(
								(i * 1.2f) - (sizeX - 1) * 0.6f,
								-2.0f,
								(j * 1.2f) - (sizeZ - 1) * 0.6f
						)
				), glm::vec3(0.5)));
			}
		}
		
		const vkcv::ImageHandle targetImage = (
				useBlurPass?
				blurBuffer :
				swapchainInput
		);
		
		core.recordDrawcallsToCmdStream(
				cmdStream,
				(showBuffer == 1?
					gBufferPosPipeline : (showBuffer == 2?
					gBufferNorPipeline : gBufferPipeline)
				),
				pushConstants,
				drawcalls,
				{
					showBuffer == 1? targetImage : positionBuffer,
					showBuffer == 2? targetImage : normalBuffer,
					showBuffer == 3? targetImage : colorBuffer,
					depthBuffer
				},
				windowHandle
		);
		
		if (((showBuffer == 0) && (useLighting)) || (showBuffer == 4)) {
			core.prepareImageForSampling(cmdStream, positionBuffer);
			core.prepareImageForSampling(cmdStream, normalBuffer);
			
			pushConstants.clear();
			
			const glm::mat4 lightMatrix = glm::rotate<float>(
					glm::identity<glm::mat4>(),
					glm::radians<float>(10.0f * static_cast<float>(t)),
					glm::vec3(0.0, 1.0, 0.0)
			);
			
			for (size_t i = 0; i < lightDrawcalls.size(); i++) {
				float x = static_cast<float>(i % 10) - 5.0f;
				float y = static_cast<float>((i / 10) % 10) - 5.0f;
				
				pushConstants.appendDrawcall(glm::translate(
						lightMatrix,
						glm::vec3(
								x * 1.2f,
								-1.0f,
								y * 1.2f
						)
				));
			}
			
			core.recordDrawcallsToCmdStream(
					cmdStream,
					lightPipeline,
					pushConstants,
					lightDrawcalls,
					{
						showBuffer == 4? targetImage : lightBuffer
					},
					windowHandle
			);
		}
		
		if (((showBuffer == 0) && (useSSAO)) || (showBuffer == 5)) {
			core.prepareImageForSampling(cmdStream, positionBuffer);
			core.prepareImageForSampling(cmdStream, normalBuffer);
			
			core.recordDrawcallsToCmdStream(
					cmdStream,
					aoPipeline,
					vkcv::pushConstants<glm::vec2>(glm::vec2(ssaoRadius, ssaoQuality)),
					{ aoDrawcall },
					{
						showBuffer == 5? targetImage : (useBlurPass? blurBuffer : aoBuffer)
					},
					windowHandle
			);
		}
		
		if (useBlurPass) {
			core.prepareImageForSampling(cmdStream, blurBuffer);
			
			core.recordDrawcallsToCmdStream(
					cmdStream,
					blurPipeline,
					vkcv::pushConstants<float>(0.0f),
					{ blurDrawcall },
					{
							showBuffer == 0? aoBuffer : swapchainInput
					},
					windowHandle
			);
		}
		
		if (showBuffer == 0) {
			core.prepareImageForSampling(cmdStream, lightBuffer);
			core.prepareImageForSampling(cmdStream, colorBuffer);
			core.prepareImageForSampling(cmdStream, aoBuffer);
			
			core.recordDrawcallsToCmdStream(
					cmdStream,
					compositePipeline,
					vkcv::pushConstants<glm::vec2>(
							glm::vec2(useLighting? 1 : 0, useSSAO? 1 : 0)
					),
					{ compositeDrawcall },
					{
							swapchainInput
					},
					windowHandle
			);
		}
		
		core.prepareSwapchainImageForPresent(cmdStream);
		core.submitCommandStream(cmdStream);
		
		gui.beginGUI();
		ImGui::Begin("Settings");
		ImGui::SliderInt("X-Axis", &sizeX, 1, 100);
		ImGui::SliderInt("Z-Axis", &sizeZ, 1, 100);
		ImGui::Spacing();
		ImGui::Combo("Show buffer", &showBuffer, showBufferNames.data(),
					 static_cast<int>(showBufferNames.size()));
		ImGui::Checkbox("Use color texture", &useColorTexture);
		ImGui::Checkbox("Use blur pass", &useBlurPass);
		ImGui::SliderFloat("SSAO radius", &ssaoRadius, 0.0f, 1.0f);
		ImGui::SliderFloat("SSAO quality", &ssaoQuality, 1.0f, 100.0f);
		ImGui::Checkbox("Composite lighting", &useLighting);
		ImGui::Checkbox("Composite SSAO", &useSSAO);
		ImGui::End();
		gui.endGUI();
	});
	
	matrixBuffer.unmap();
	texturePropsBuffer.unmap();
	resolutionBuffer.unmap();
	return 0;
}
